﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using anmar.SharpMimeTools;
using EricDaugherty.CSES;
using EricDaugherty.CSES.Common;
using EricDaugherty.CSES.SmtpServer;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.ServiceRuntime;
using Microsoft.WindowsAzure.StorageClient;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections.Specialized;
using System.Net;
using System.Net.Mail;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;

namespace EmailTheInternet_SMTP
{
    public class Utility
    {
        public static string GetCdnUrlForUri(Uri uri)
        {
            var builder = new UriBuilder(uri);
            builder.Host = RoleEnvironment.GetConfigurationSettingValue("CdnHostName");
            return builder.Uri.AbsoluteUri;
        }
    }

    public class RecipientFilter : IRecipientFilter
    {
        public bool AcceptRecipient(SMTPContext context, EmailAddress recipient)
        {
            return recipient.Username.ToLower() == "post";
        }
    }

    public class MessageSpool : IMessageSpool
    {
        static public void SaveBody(SharpMessage msg, CloudBlob bodyBlob, SMTPMessage originalMessage, CloudBlobContainer container, string permalink)
        {
            var body = msg.Body;
            int where = body.Length;
            string footerFormat;
            string additionalAttachmentsFormat;
            string tagFormat;
            string attachmentSeparator;
            string attachmentFormat;
            if (msg.HasHtmlBody)
            {
                bodyBlob.Properties.ContentType = "text/html";
                var closeBodyTags = Regex.Matches(body, @"<\s*/\s*body\s*>");
                if (closeBodyTags.Count > 0)
                {
                    where = closeBodyTags[closeBodyTags.Count - 1].Index;
                }
                footerFormat = "<div style=\"clear:both\">\r\n\t<hr />";
                additionalAttachmentsFormat = "\r\n\tAdditional attachments:\r\n\t<ul>\r\n\t\t{0}\r\n\t</ul>";
                attachmentFormat = @"<li><a href=""{0}"">{1}</a></li>";
                tagFormat = "\r\n\t<div style=\"text-align:right\"><a href=\"http://{0}\">{0}</a></div>\r\n</div>";
                attachmentSeparator = "\r\n\t\t";
            }
            else
            {
                bodyBlob.Properties.ContentType = "text/plain";
                footerFormat = "\r\n\r\n---";
                additionalAttachmentsFormat = "\r\nAdditional attachments:\r\n{0}";
                attachmentSeparator = "\r\n";
                attachmentFormat = "* {0}";
                tagFormat = "\r\n\r\nhttp://{0}";
            }

            // get the attachments that aren't referenced inline (so we can link to them)
            var additionalAttachments = msg.Attachments.Cast<SharpAttachment>().Where(a => !a.Inline).Select(a => string.Format(attachmentFormat, Utility.GetCdnUrlForUri(container.GetBlobReference(permalink + "/" + a.Name).Uri), a.Name)).ToArray();
            body = body.Insert(where, string.Format(footerFormat + (additionalAttachments.Length > 0 ? additionalAttachmentsFormat : "") + string.Format(tagFormat, RoleEnvironment.GetConfigurationSettingValue("DomainName")),
                string.Join(attachmentSeparator, additionalAttachments)
                ));
            try
            {
                // this is pretty ugly, but it seems that SharpMessage() doesn't expose the character set used
                // if we don't do this, email from GMail gets uploaded incorrectly (notably "=A0", which should be a space, gets mangled)
                var msg2 = new SharpMimeMessage(new MemoryStream(Encoding.ASCII.GetBytes(originalMessage.Data)));
                var charset = msg2.GetPart(msg2.PartsCount - 1).Header.ContentTypeParameters["charset"];
                // try to use the encoding found via content-type charset
                bodyBlob.UploadText(body, Encoding.GetEncoding(charset), null);
            }
            catch
            {
                // fall back to a default encoding
                bodyBlob.UploadText(body);
            }
        }

        static public string ConstructReply(SharpMessage msg, CloudBlob bodyBlob)
        {
            var body = msg.Body;
            int where = 0;
            string responseFormat, originalMessageSeparator, headerSeparator, headerFormat, headerEnd;
            if (msg.HasHtmlBody)
            {
                var bodyTags = Regex.Matches(body, @"<\s*body[^>]*>");
                if (bodyTags.Count > 0)
                {
                    var match = bodyTags[bodyTags.Count - 1];
                    where = match.Index + match.Length;
                }
                responseFormat = @"Thanks!  You can find your content here: <a href=""{0}"">{0}</a>";
                originalMessageSeparator = "<hr />";
                headerSeparator = "<br />";
                headerFormat = "<b>{0}</b> {1}";
                headerEnd = "<br /><br />";
            }
            else
            {
                responseFormat = @"Thanks!  You can find your content here: {0}.";
                originalMessageSeparator = "\r\n\r\n-----Original Message-----\r\n";
                headerSeparator = "\r\n";
                headerFormat = "{0}: {1}";
                headerEnd = "\r\n\r\n";
            }
            var headers = new OrderedDictionary();
            headers.Add("To:", msg.From);
            headers.Add("Sent:", msg.Date.ToString("dddd, MMMM dd, yyyy h:mm tt"));
            headers.Add("From:", RoleEnvironment.GetConfigurationSettingValue("EmailAddress"));
            headers.Add("Cc:", string.Join("; ", msg.Cc.Cast<SharpMimeAddress>().Select(c => c.ToString()).ToArray()));
            headers.Add("Subject:", msg.Subject);
            body = body.Insert(where, string.Format(responseFormat, Utility.GetCdnUrlForUri(bodyBlob.Uri)) + originalMessageSeparator +
                string.Join(headerSeparator, headers.Keys.Cast<string>().Where(k => !string.IsNullOrEmpty((string)headers[k])).Select(k => string.Format(headerFormat, k, headers[k])).ToArray()) +
                headerEnd);
            return body;
        }

        public bool SpoolMessage(SMTPMessage message)
        {
            try
            {
                var account = CloudStorageAccount.Parse(RoleEnvironment.GetConfigurationSettingValue("DataConnectionString"));

                // make a container, with public access to blobs
                var id = Guid.NewGuid().ToString().Replace("-", null);
                var container = account.CreateCloudBlobClient().GetContainerReference(id);
                container.Create();
                container.SetPermissions(new BlobContainerPermissions() { PublicAccess = BlobContainerPublicAccessType.Blob });

                // parse the message
                var msg = new SharpMessage(new MemoryStream(Encoding.ASCII.GetBytes(message.Data)), SharpDecodeOptions.AllowAttachments | SharpDecodeOptions.AllowHtml | SharpDecodeOptions.DecodeTnef);

                // create a permalink-style name for the blob
                var permalink = Regex.Replace(Regex.Replace(msg.Subject.ToLower(), @"[^a-z0-9]", "-"), "--+", "-").Trim('-');
                if (string.IsNullOrEmpty(permalink))
                {
                    // in case there's no subject
                    permalink = "message";
                }
                var bodyBlob = container.GetBlobReference(permalink);
                // set the CDN to cache the object for 2 hours
                bodyBlob.Properties.CacheControl = "max-age=7200";

                // replaces references to attachments with the URL of where we'll put them
                msg.SetUrlBase(Utility.GetCdnUrlForUri(bodyBlob.Uri) + "/[Name]");

                // save each attachment in a blob, setting the appropriate content type
                foreach (SharpAttachment attachment in msg.Attachments)
                {
                    var blob = container.GetBlobReference(permalink + "/" + attachment.Name);
                    blob.Properties.ContentType = attachment.MimeTopLevelMediaType + "/" + attachment.MimeMediaSubType;
                    blob.Properties.CacheControl = "max-age=7200";
                    attachment.Stream.Position = 0;
                    blob.UploadFromStream(attachment.Stream);
                }

                // add the footer and save the body to the blob
                SaveBody(msg, bodyBlob, message, container, permalink);

                // construct the body of our reply
                var body = ConstructReply(msg, bodyBlob);

                // actually sending the email is the easy part
                var reply = new MailMessage(RoleEnvironment.GetConfigurationSettingValue("EmailAddress"), msg.FromAddress)
                {
                    Subject = msg.Subject.StartsWith("RE:", StringComparison.InvariantCultureIgnoreCase) ? msg.Subject : string.Format("RE: " + msg.Subject),
                    Body = body,
                    IsBodyHtml = msg.HasHtmlBody // send HTML if we got HTML
                };
                if (!reply.IsBodyHtml) reply.BodyEncoding = Encoding.UTF8;
                // make it a proper reply
                reply.Headers["References"] = msg.MessageID;
                reply.Headers["In-Reply-To"] = msg.MessageID;
                // use our SMTP server, port, username, and password to send the mail
                (new SmtpClient(RoleEnvironment.GetConfigurationSettingValue("SmtpServer"), int.Parse(RoleEnvironment.GetConfigurationSettingValue("SmtpPort")))
                {
                    Credentials = new NetworkCredential(RoleEnvironment.GetConfigurationSettingValue("SmtpUsername"), RoleEnvironment.GetConfigurationSettingValue("SmtpPassword"))
                }).Send(reply);

                return true;
            }
            catch (Exception ex)
            {
                // log the exception and return false so we reject the message
                Debug.WriteLine(ex.ToString());
                return false;
            }
        }
    }

    public class SMTPHandler
    {
        private TcpListener listener;
        private SMTPProcessor processor;

        public SMTPHandler()
        {
            listener = new TcpListener(IPAddress.Any, RoleEnvironment.CurrentRoleInstance.InstanceEndpoints["SmtpIn"].IPEndpoint.Port);
            processor = new SMTPProcessor(RoleEnvironment.GetConfigurationSettingValue("DomainName"), new RecipientFilter(), new MessageSpool());
            listener.Start();
        }

        public void Run()
        {
            var mutex = new ManualResetEvent(false);
            while (true)
            {
                mutex.Reset();
                listener.BeginAcceptSocket((ar) =>
                    {
                        mutex.Set();
                        processor.ProcessConnection(listener.EndAcceptSocket(ar));
                    }, null);
                mutex.WaitOne();
            }
        }
    }
}